/*
 *
 * Copyright (C) 2000  Robert Olsson, Jens Ls 
 * Swedish University of Agricultural Sciences
 *
 * This file is part of GNU Zebra.
 *
 * GNU Zebra is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * GNU Zebra is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with GNU Zebra; see the file COPYING.  If not, write to the Free
 * Software Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  
 */

#include <zebra.h>
#include "vector.h"
#include "vty.h"
#include "command.h"
#include "getopt.h"
#include "thread.h"
#include "stream.h"
#include "network.h"
#include "prefix.h"
#include "version.h"
#include "memory.h"
#include "if.h"
#include "sockunion.h"
#include "log.h"
#include "pim.h"
#include "pimd.h"
#include "pim_fsm.h"
#include <string.h>
#include <netinet/igmp.h>
#include <linux/mroute.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include "assert.h"
#include "tables.h"
#include "pim_route.h"
#include "pim_packet.h"
#include "igmp_def.h"
#include "ioctl.h"
#include "if_idx_nexthop.h"
#include "pim_joinprune.h"
#include "pim_assert.h"


#include "zbug.h"
ZBUG_EXTERN

extern int pim_sock;
extern int in_cksum (void *ptr, int nbytes);
extern list pim_neighbor_list;
extern list mrt_hash[];
extern list mgrp_hash[];
extern list msrc_hash[MSRC_HASH_SIZE];
extern struct best_route best_route;
extern int pim_sock, igmp_sock;
extern char pim_sendbuf[];


char b1[16], b2[16], b3[16], b4[16];  /* For inet_2a */



u_int32_t mg_get_metric(struct mgrp *mg, struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;

  if(mg->flags & MRT_ASSERTED) 
    return mg->pim.asserted.metric;

  /* If we have no metric for this (*,G)  we use the configured interface
     metrics for the interfacw */
    
  if( mg->pim.rp->metric == ASSERT_NOTFOUND_METRIC )
    return pi->metric;
  else 
    return  mg->pim.rp->metric;
}

u_int32_t mg_get_preference(struct mgrp *mg, struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;

  if(mg->flags & MRT_ASSERTED) 
    return mg->pim.asserted.preference;

  /* If we have no preference for this (*,G) we use the configured interface
     preferences for the interfacw */
    
  if( mg->pim.rp->preference == ASSERT_NOTFOUND_PREFERENCE )
    return pi->preference;
  else 
    return  mg->pim.rp->preference;
}


u_int32_t mr_get_metric(struct mrt *mr, struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;

  if(mr->flags & MRT_ASSERTED) 
    return mr->asserted.metric;

  /* If we have no metric for this source we use the configured interface
     metrics for the interfacw */
    
  if( mr->ms->metric == ASSERT_NOTFOUND_METRIC )
    return pi->metric;
  else 
    return  mr->ms->metric;
}

u_int32_t mr_get_preference(struct mrt *mr, struct interface *ifp)
{
  struct pim_interface *pi = ifp->info;

  if(mr->flags & MRT_ASSERTED) 
    return mr->asserted.preference;

  /* If we have no preference for this source we use the configured interface
     preferences for the interfacw */
    
  if( mr->ms->preference == ASSERT_NOTFOUND_PREFERENCE )
    return pi->preference;
  else 
    return  mr->ms->preference;
}



  /* Test is assert wins over current perference/metric
     then return TRUE

     The order for comparison is 
     1) RPT.  2) Preference. 3) Metric 

     And assserts in action has to checked 
     */


int
assert_wins(  struct mgrp *mg, 
	      struct mrt *mr, 
	      u_int32_t rpt,
	      u_int32_t preference,
	      u_int32_t metric,
	      struct interface *ifp,
	      u_int32_t src)

{
  struct pim_interface *pi = ifp->info;
  u_int32_t our_preference, our_metric;


  if(rpt ) {

    our_metric = mg_get_metric(mg, ifp);
    our_preference = mg_get_preference(mg, ifp);
    
    if( our_preference > preference ) return 1;
    if( our_preference < preference ) return 0;
    
    if( our_metric > metric ) return 1;
    if( our_metric < metric) return 0;

    /* Tie. Select higher IP-address */
  
    if(ntohl(src) > ntohl(pi->address->u.prefix4.s_addr)) return 1;
    return 0;
  }  

  if(mr) {

    
    our_metric = mr_get_metric(mr, ifp);
    our_preference = mr_get_preference(mr, ifp);
        
    if( our_preference > preference ) return 1;
    if( our_preference < preference ) return 0;
    
    if( our_metric > metric ) return 1;
    if( our_metric < metric ) return 0;
  }  
  
  /* Tie. Select higher IP-address */
  
  if(ntohl(src) > ntohl(pi->address->u.prefix4.s_addr)) return 1;
  return 0;

}


void
send_pim_assert(struct interface *ifp, 
		u_int32_t group, 
		u_int32_t source, 
		u_int32_t rpt,
		u_int32_t preference, 
		u_int32_t metric)
{
 struct pim_interface *pi = ifp->info;
 void *buf_p, *p;
 u_int32_t len, tmp;
 time_t now;

 p = buf_p = (char *) (pim_sendbuf + sizeof(struct ip) 
   + sizeof(struct pimheader));

 buf_p += encode_group(buf_p, group, 32);
 buf_p += encode_unicast(buf_p, source);

 tmp = htonl(preference);
 if(rpt) tmp |= htonl(ASSERT_RPT_BIT);
 memcpy(buf_p, &tmp, 4);
 buf_p += 4;
 tmp = htonl(metric);
 memcpy(buf_p, &tmp, 4);
 buf_p += 4;
 
 len = buf_p - p;

 printf("TX Assert (%s, %s) pref/metric [%x/%x] on %s\n",
	(rpt)? "*" : inet_2a(source, b1),
	inet_2a(group, b2), preference, metric, ifp->name);

  ZBUG_G(group)
      {
	time(&now);
	
	fprintf(ZBUG_LOG,"TX Assert (%s, %s) pref/metric [%x/%x] on %s %s",
		(rpt)? "*" : inet_2a(source, b1),
		inet_2a(group, b2),
		preference,
		metric,
		ifp->name,
		ctime(&now));
	fflush(ZBUG_LOG);
      }


 if(pi->version == 2)
 send_pim_v2_packet(ifp, pim_sendbuf, len, 
                    PIM_ASSERT, htonl(INADDR_PIM_ALLRTRS_GROUP), 1);

}


void
process_assert_outgoing(struct encoded_group *eg, 
			struct encoded_unicast *eu, 
			u_int32_t rpt,
			u_int32_t preference,
			u_int32_t metric,
			struct interface *ifp,
			u_int32_t src,
			struct mgrp *mg,
			struct mrt *mr)
{
  struct pim_interface *pi = ifp->info;
  struct msrc *ms;
  vifbitmap_t oifs;

  
  /* 3.5.2.1 (1)  */
  
  if(!mr && !rpt) {
    
    /* Add source if needed */
    
    if((ms= ms_lookup( eu->addr)) == NULL ) {
      ms = ms_add( eu->addr);
      if(!ms) return;
    }

    mr = mr_add(ms, mg);
    mr->flags |= 
      (MRT_SPARSE|MRT_RPT|MRT_PIM_IN_USE); 
    
    copy_mg_oifs(mr, mg);
    
    mr_entry_timer_update(mr, pi->data_timeout);
    
    oifs = mr_oifs(mr);
    mfc_add(igmp_sock,
	    (u_int32_t) eu->addr,
	    (u_int32_t) eg->addr,
	    mr_iif( mr ), 
	    oifs);
  }

  if(!assert_wins(mg, mr, rpt, preference, metric, ifp, src ))
    {
      u_int32_t our_metric, our_preference;


      /* We won: Keep oif and send assert */
      if(rpt) {
	our_metric = mg_get_metric(mg, ifp);
	our_preference = mg_get_preference(mg, ifp);

	send_pim_assert(ifp, 
			mg->group, 
			eu->addr,
			rpt,
			our_preference, 
			our_metric);
      }
      else {
	our_metric = mr_get_metric(mr, ifp);
	our_preference = mr_get_preference(mr, ifp);
	send_pim_assert(ifp, 
			mg->group, 
			eu->addr,
			rpt,
			our_preference, 
			our_metric);
      }
    }
  else
    {
      /* We lost: Delete oif from entry */
      if(rpt)
	{
	  /* Delete from (*,G) and all associated (S,G) */
	  /* In reality we set the MRT_ASSERTED flag on the oifs */
	  mg_flag_oif(mg, ifp, MRT_ASSERTED);

	  printf("RPT we lost \n");
	  
	  /* mfc_add() each S,G entry */
	  mg_mfc_add(mg);
	}
      else
	{
	  /* Delete from (S,G) */
	  /* In reality we set the MRT_ASSERTED flag on the oif */
	  mr_flag_oif(mr, ifp, MRT_ASSERTED);

	  printf("We lost \n");

	  /* mfc_add (S,G) */
	  mr_mfc_add(mr);
	}
      /*
	THINK: When ASSERT flips back so we get WRONGVIFs again
	we already have the oif, although it is marked with MRT_ASSERTED */
    }

}

void
process_assert_incoming(struct encoded_group *eg, 
			struct encoded_unicast *eu, 
			u_int32_t rpt,
			u_int32_t preference,
			u_int32_t metric,
			struct interface *ifp,
			u_int32_t src,
			struct mgrp *mg,
			struct mrt *mr)
{
  listnode node;
  
  if( ! mr ) return; /* TODO:  Check this */
  
  /* 1 */
  if(rpt)
    {
      if(mg->flags & MRT_ASSERTED)
	if(!assert_wins(mg, mr, rpt, preference, metric, ifp, src ))
	  return;
    }
  else
    {
      if(mr->flags & MRT_ASSERTED)
	if(!assert_wins(mg, mr, rpt, preference, metric, ifp, src ))
	  return;
    }


  if(rpt)
    {
      /* Switch rpf_neighbor for (*,G) */
      if(mg->pim.t_assert)
	thread_cancel(mg->pim.t_assert);
      
      if(mg->pim.rp->rpf_neighbor == src)
	{
	  mg->flags &= ~MRT_ASSERTED;
	  for(node=listhead(mg->mrt_list); node; nextnode(node))
	    {
	      mr = getdata( node );
	      if(!(mr->flags & MRT_SPT))
		mr->flags &= ~MRT_ASSERTED;
	      if(mr->flags & MRT_REGISTER)
		mr_mfc_add(mr);/* start sending register */
	    }
	}
      else
	{
	  mg->flags |= MRT_ASSERTED;
	  mg->pim.asserted.preference = preference; 
	  mg->pim.asserted.metric = metric;
	  mg->pim.asserted.rpf_neighbor = src;
      
	  /* switch rpf neighbor for all non SPT (S,G)?? */
	  for(node=listhead(mg->mrt_list); node; nextnode(node))
	    {
	      mr = getdata( node );
	      if(!(mr->flags & MRT_SPT))
		{
		  mr->flags |= MRT_ASSERTED;
		  mr->asserted.preference = preference; 
		  mr->asserted.metric = metric;
		  mr->asserted.rpf_neighbor = src;
		}
	      if(mr->flags & MRT_REGISTER)
		mr_mfc_add(mr); /* stop sending register */
	    }
	  //    mg->pim.t_assert = thread_add_timer(master, pim_mg_assert_timeout, mg, 66);
	}
    }
  else
    {
      /* Switch rpf_neighbor for (S,G) */
      if(mr->t_assert)
	thread_cancel(mr->t_assert);
      
      mr->flags |= MRT_ASSERTED;
      mr->asserted.preference = preference; 
      mr->asserted.metric = metric;
      mr->asserted.rpf_neighbor = src;
   //   mr->t_assert = thread_add_timer(master, pim_mr_assert_timeout, mr, 66);

    }
  
  /* 2 */
  if(rpt)
    {
      if(! list_isempty( mg->oif_list ))
	{
	  /* We have downstream members */
	  trigger_join_prune( ifp );
	  
	}
    }
  else
    {
      if(! list_isempty( mr->oif_list ))
	{
	  /* We have downstream members */
	  trigger_join_prune( ifp );
	  
	}
    }
}

void
process_assert(struct encoded_group *eg, 
	       struct encoded_unicast *eu, 
	       u_int32_t rpt,
	       u_int32_t preference,
	       u_int32_t metric,
	       struct interface *ifp,
	       u_int32_t src)
{
  struct mgrp *mg;
  struct mrt *mr = 0;

  mg = mg_lookup(eg->addr);
  mr = mr_lookup(eu->addr, eg->addr);

  printf("RX Assert for (%s, %s)  %s preference=%x, metric=%x\n", 
	 inet_2a(eu->addr, b1), 
	 inet_2a(eg->addr, b2), 
	 rpt? "RPT-bit":"", preference, metric);

  /* Decide if this is received on an outgoing or incoming interface */

  /* 3.5.2 (1) pp 21 */
  /*  if(rpt && !mg ) return;*/
  if( !mg ) return;

  if(!(mg->flags & MRT_PIM_IN_USE)) return ; /* PIM active  ? */
  
  if(rpt) {
    if(mg->pim.rp->ifp == ifp )  
      process_assert_incoming(eg, eu, rpt, preference,  metric, 
			      ifp, src, mg, mr);
    else 
      process_assert_outgoing(eg, eu, rpt, preference, metric, 
			      ifp, src, mg, mr);
    return ;
  }
  
  /* rpt is not set. Match  (S,G) first or (*,G) */

  /* 3.5.2 (2) pp 21 */

  

  if(mr && (mr->flags & MRT_SPT)) {
    if(mr->ms->ifp == ifp) 
      process_assert_incoming(eg, eu, rpt, preference,  metric, 
			      ifp, src, mg, mr);
    else 
      process_assert_outgoing(eg, eu, rpt, preference,   metric, 
			      ifp, src, mg, mr);
    return ;
  }
}


DEFUN (if_ip_pim_preference,
       if_ip_pim_preference_cmd,
       "ip pim preference <0-65535>",
       "PIM interface commands\n"
       "Specify PIM preference\n"
       "Specify PIM preference\n")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;

  pi->flags |= IF_PREFERENCE_SET;
  pi->preference = strtol (argv[0], NULL, 10);

  return CMD_SUCCESS;
}

DEFUN (if_no_ip_pim_preference,
       if_no_ip_pim_preference_cmd,
       "no ip pim preference <0-65535>",
       "PIM interface default preference\n"
       "PIM interface default preference")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;

  pi->flags &= ~IF_PREFERENCE_SET;
  pi->preference = ASSERT_DEFAULT_PREFERENCE;

  return CMD_SUCCESS;
}


DEFUN (if_ip_pim_metric,
       if_ip_pim_metric_cmd,
       "ip pim metric <0-65535>",
       "PIM interface commands\n"
       "Specify PIM metric\n"
       "Specify PIM metric\n")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;

  pi->flags |= IF_METRIC_SET;
  pi->metric = strtol (argv[0], NULL, 10);

  return CMD_SUCCESS;
}

DEFUN (if_no_ip_pim_metric,
       if_no_ip_pim_metric_cmd,
       "no ip pim metric <0-65535>",
       "PIM interface default metric\n"
       "PIM interface default metric")
{
  struct interface *ifp;
  struct pim_interface *pi;

  ifp = vty->index;
  pi = ifp->info;

  pi->flags &= ~IF_METRIC_SET;
  pi->metric = ASSERT_DEFAULT_METRIC;

  return CMD_SUCCESS;
}

void
pim_assert_init()
{
  install_element (INTERFACE_NODE, &if_ip_pim_preference_cmd);
  install_element (INTERFACE_NODE, &if_no_ip_pim_preference_cmd);

  install_element (INTERFACE_NODE, &if_ip_pim_metric_cmd);
  install_element (INTERFACE_NODE, &if_no_ip_pim_metric_cmd);
}
